#include "lusol.h"

/*-----------------------------------------------*/
int cuda_init() {
  int deviceCount, dev;
  cudaGetDeviceCount(&deviceCount);
  printf("=========================================\n");
  if (deviceCount == 0) {
    printf("There is no device supporting CUDA\n");
    return 0;
  }
  for (dev = 0; dev < deviceCount; ++dev) {
    cudaDeviceProp deviceProp;
    cudaGetDeviceProperties(&deviceProp, dev);
    if (dev == 0) {
      if (deviceProp.major == 9999 && deviceProp.minor == 9999)
        printf("There is no device supporting CUDA.\n");
      else if (deviceCount == 1)
        printf("There is 1 device supporting CUDA\n");
      else
        printf("There are %d devices supporting CUDA\n", deviceCount);
    }
    printf("\nDevice %d: \"%s\"\n", dev, deviceProp.name);
    printf("  Major revision number:          %d\n",
           deviceProp.major);
    printf("  Minor revision number:          %d\n",
           deviceProp.minor);
    printf("  Total amount of global memory:  %.2f GB\n",
           deviceProp.totalGlobalMem/1e9);
  }
  dev = 0;
  cudaSetDevice(dev);
  cudaDeviceProp deviceProp;
  cudaGetDeviceProperties(&deviceProp, dev);
  printf("\nRunning on Device %d: \"%s\"\n", dev, deviceProp.name);
  printf("=========================================\n");
  return 1;
}

/*---------------------------------------------------*/
void cuda_check_err() {
  cudaError_t cudaerr = cudaGetLastError() ;
  if (cudaerr != cudaSuccess) 
    printf("error: %s\n",cudaGetErrorString(cudaerr));
}

/*----------------------------------------------------------*/
void cudaMallocCSR(int n, int nnz, struct csr_t *d_csr)
{
  cudaMalloc((void **)&d_csr->ia, (size_t)(n+1)*sizeof(int));
  cudaMalloc((void **)&d_csr->ja, (size_t)(nnz)*sizeof(int));
  cudaMalloc((void **)&d_csr->a,  (size_t)(nnz)*sizeof(REAL));
}

/*----------------------------------------------------------*/
void CSRHost2Device(struct csr_t *h_csr, struct csr_t *d_csr)
{
  int n = h_csr->n;
  int nnz = h_csr->nnz;

  cudaMemcpy(d_csr->ia, h_csr->ia, (size_t)(n+1)*sizeof(int),
             cudaMemcpyHostToDevice);
  cudaMemcpy(d_csr->ja, h_csr->ja, (size_t)(nnz)*sizeof(int),
             cudaMemcpyHostToDevice);
  cudaMemcpy(d_csr->a,  h_csr->a,  (size_t)(nnz)*sizeof(REAL),
             cudaMemcpyHostToDevice);
}

/*-------------------------------------------------------------*/
void cudaMallocLU(int n, int nnzl, int nnzu, struct lu_t *d_lu)
{
  CUDA_SAFE_CALL(cudaMalloc((void **)&d_lu->l.ia, (size_t)(n+1)*sizeof(int)));
  CUDA_SAFE_CALL(cudaMalloc((void **)&d_lu->l.ja, (size_t)(nnzl)*sizeof(int)));
  CUDA_SAFE_CALL(cudaMalloc((void **)&d_lu->l.a,  (size_t)(nnzl)*sizeof(REAL)));

  CUDA_SAFE_CALL(cudaMalloc((void **)&d_lu->u.ia, (size_t)(n+1)*sizeof(int)));
  CUDA_SAFE_CALL(cudaMalloc((void **)&d_lu->u.ja, (size_t)(nnzu)*sizeof(int)));
  CUDA_SAFE_CALL(cudaMalloc((void **)&d_lu->u.a,  (size_t)(nnzu)*sizeof(REAL)));
}

/*------------------------------------------------------------*/
void LUHost2Device(struct lu_t *h_lu, struct lu_t *d_lu)
{
  // L
  int n   = h_lu->l.n;
  int nnz = h_lu->l.nnz;
  CUDA_SAFE_CALL(cudaMemcpy(d_lu->l.ia, h_lu->l.ia, (size_t)(n+1)*sizeof(int),
             cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(d_lu->l.ja, h_lu->l.ja, (size_t)(nnz)*sizeof(int),
             cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(d_lu->l.a,  h_lu->l.a,  (size_t)(nnz)*sizeof(REAL),
             cudaMemcpyHostToDevice));

  // U
  n   = h_lu->u.n;
  nnz = h_lu->u.nnz;

  CUDA_SAFE_CALL(cudaMemcpy(d_lu->u.ia, h_lu->u.ia, (size_t)(n+1)*sizeof(int),
             cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(d_lu->u.ja, h_lu->u.ja, (size_t)(nnz)*sizeof(int),
             cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(d_lu->u.a,  h_lu->u.a,  (size_t)(nnz)*sizeof(REAL),
             cudaMemcpyHostToDevice));
}

/*---------------------------------*/
void CudaFreeCSR(struct csr_t *d_csr)
{
  cudaFree(d_csr->ia);
  cudaFree(d_csr->ja);
  cudaFree(d_csr->a);
}

/*------------------------------*/
void CudaFreeLU(struct lu_t *d_lu)
{
  cudaFree(d_lu->l.ia);
  cudaFree(d_lu->l.ja);
  cudaFree(d_lu->l.a);
  cudaFree(d_lu->u.ia);
  cudaFree(d_lu->u.ja);
  cudaFree(d_lu->u.a);
}

/*-----------------------------------*/
void CudaFreeLev(struct level_t *d_lev)
{
  cudaFree(d_lev->jlevL);
  cudaFree(d_lev->ilevL);
  cudaFree(d_lev->jlevU);
  cudaFree(d_lev->ilevU);
}

